#!/usr/bin/env python
#_*_coding:utf-8_*_
# Authors: Superzchen

import fileinput, re, os, sys, Protein.index_list, pickle
from os import listdir
from os.path import isfile, join
import numpy as np
from RNA import *
from DNA.util import *
from Protein.extract_aaindex import norm_index_vals, AAIndex

def readFasta(file):
	if os.path.exists(file) == False:
		print('Error: "' + file + '" does not exist.')
		sys.exit(1)

	with open(file) as f:
		records = f.read()

	if re.search('>', records) == None:
		print('The input file seems not in fasta format.')
		sys.exit(1)

	records = records.replace('\r','').split('>')[1:]
	myFasta = []
	for fasta in records:
		array = fasta.split('\n')
		name, sequence = array[0].split()[0], re.sub('[^ARNDCQEGHILKMFPSTWYVU-]', '-', ''.join(array[1:]).upper())
		myFasta.append([name, sequence])
	return myFasta

class SeqSS:
    """An object that contains RNA sequence and its secondary structure"""
    def __init__(self, id, sequence, sstructure, no, mfe):
        self.id = id
        self.sequence = sequence.upper()
        self.sstructure = sstructure            #its secondary structure
        self.no = no
        self.length = len(sequence) if len(sequence)==len(sstructure) else -1
        self.MFE = mfe          #float

    def __str__(self):
        """Output seqSS when 'print' method is called."""
        return "%s\tNo:%s\tlength:%s\tMFE:%.2f\n%s\n%s" % (self.id, str(self.no), str(self.length), self.MFE, self.sequence, self.sstructure)

def get_rnasc_data(input_data):
    """Get rnasc data from file or list with check.

    :param input_data: type file or list
    :param desc: with this option, the return value will be a SeqSS object list(it only works in file object).
    :return: a SeqSS object or shutdown.
    """
    if hasattr(input_data, 'read'):
        return read_rnasc_fasta(input_data)
    elif isinstance(input_data, list):
        input_data = is_rnasc_list(input_data)
        if input_data is not False:
            return input_data
        else:
            sys.exit(0)
    else:
        error_info = 'Sorry, the parameter in get_rnasc_data method must be list or file type.'
        sys.exit(error_info)

def is_rnasc_list(rnasc_list): 
    """Judge the rnasc_list is within the scope of alphabet and change the lowercase to capital.
    The format of the input list is as follows:
        [">sequence id","RNA sequence should be consist of AGCU","second structure"]"""
    seqss_list = []
    count = 0    # record line numbers
    no = 0       # record sequence numbers
    alphabet = index_list.RNA
    id,sequence,sstruc = '','',''
    seqss_id = r'>(.+)'
    seqss_sequence = r'([ACUGacug]+)'
    seqss_sstructure = r'([\(\)\.]+)'
    seqss_MFE = r'-\d+\.\d+'
    mfe = ""
    for line in rnasc_list:
            
        if not line:
            break
            
        if count %3 == 0:
            id = re.findall(seqss_id, line)[0]
        elif count%3 == 1:
            res = is_under_alphabet(line, alphabet)
            if res is not True:
                error_info = 'Error, sequence ' + str(no) \
                             + ' has character ' + str(res) + '.(The character must be ' + alphabet + ').'
                sys.stderr.write(error_info)
                return False
            else:
                sequence = re.findall(seqss_sequence, line)[0]
                
        elif count%3 == 2:
            sstruc = re.findall(seqss_sstructure, line)[0]
            if re.findall(seqss_MFE, line):
                mfe = float(re.findall(seqss_MFE, line)[0])
            else:
                mfe = 0                        #MFE is no match.
            if sequence != "" and sstruc != "":
                if is_rnasc_fasta(SeqSS(id, sequence, sstruc, no, mfe)):
                    seqss_list.append(SeqSS(id,sequence,sstruc,no, mfe))
                    no += 1
                else:
                    sys.exit()    
        count += 1
    return seqss_list
    
    
def read_rnasc_fasta(f):
    """Read a fasta file.
    
    :param f: HANDLE to input. e.g. sys.stdin, or open(<file>)
    
    Return SeqSS obj list.
    """
    seqss_list = []
    count = 0    # record line numbers
    no = 0       # record sequence numbers
    id,sequence,sstruc = '','',''
    seqss_id = r'>(.+)'
    seqss_sequence = r'([ACUG]+)'
    seqss_sstructure = r'([\(\)\.]+)'
    seqss_MFE = r'-\d+\.\d+'
    mfe = ""
    for line in f:
            
        if not line:
            break
            
        if count %3 == 0:
            id = re.findall(seqss_id, line)[0]
        elif count%3 == 1:
            sequence = re.findall(seqss_sequence, line)[0]
        elif count%3 == 2:
            sstruc = re.findall(seqss_sstructure, line)[0]
            if re.findall(seqss_MFE, line):
                mfe = float(re.findall(seqss_MFE, line)[0])
            else:
                mfe = 0  #MFE is no match.
            if is_rnasc_fasta(SeqSS(id, sequence, sstruc, no, mfe)):
                seqss_list.append(SeqSS(id,sequence,sstruc,no, mfe))
                no += 1
            else:
                sys.exit()
        count += 1
    return seqss_list

def is_rnasc_fasta(seqss):
    """Judge the SeqSS object is in FASTA format.
    Five situations will:
    1. No seqss id.
    2. SeqSS id is illegal.
    3. No sequence.
    4. The length of sequence is not equal to the length of its secondary structure
    5. The RNA sequence was not composed by AUCG
    :param seqss: SeqSS object.
    """
    alphabet ="AUCG"
    if not seqss.id:
        error_info = 'Error, sequence ' + str(seqss.no) + ' has no sequence id.'
        sys.stderr.write(error_info)
        return False
    if 0 == seqss.length:
        error_info = 'Error, sequence ' + str(seqss.no) + ' is null.'
        sys.stderr.write(error_info)
        return False
    if len(seqss.sequence) != len(seqss.sstructure):
        error_info = 'Error, the length of sequence ' + str(seqss.no) + ' is not equal to the length of its secondary structure.'
        sys.stderr.write(error_info)
        return False
        
    res = is_under_alphabet(seqss.sequence, alphabet)
    if res is not True:
        error_info = 'Error, sequence ' + str(seqss.no) \
                         + ' has character ' + str(res) + '.(The character must be ' + alphabet + ').'
        sys.stderr.write(error_info)
        return False
                         
    return True

def is_under_alphabet(s, alphabet):
    """Judge the string is within the scope of the alphabet or not.

    :param s: The string.
    :param alphabet: alphabet.

    Return True or the error character.
    """
    for e in s:
        if e not in alphabet:
            return e

    return True

def checkPSSM(inputFile, pssmdir):
    """
    Authors: Chris (chris@nohup.cc) & Young (young@nohup.cc)
    """
    check_head = re.compile(r'\>')

    smplist = []
    smpcnt = 0
    for line, strin in enumerate(fileinput.input(inputFile)):
        if not check_head.match(strin):
            smplist.append(strin.strip())
            smpcnt += 1
    #print "smplist="
    #print smplist
    onlyfiles = [ f for f in listdir(pssmdir) if isfile(join(pssmdir,f)) ]
    #print "onlyfiles="
    #print onlyfiles
    fastaDict = {}

    for fi in onlyfiles:
        cntnt = ''
        pssmContentMatrix=readToMatrix(fileinput.input(pssmdir+'/'+fi))
        pssmContentMatrix=np.array(pssmContentMatrix)
        sequence=pssmContentMatrix[:,0]
        seqLength=len(sequence)
        for i in range(seqLength):
            cntnt+=sequence[i]
        if cntnt in fastaDict:
            #print strin
            continue
        fastaDict[cntnt] = fi

    finalist = []
    for smp in smplist: 
        #print "smp="+smp
        #print "fastaDict[smp]="+fastaDict[smp]
        finalist.append(pssmdir+'/'+fastaDict[smp])
    return finalist

def readToMatrix(input_matrix):
    """
    Authors: Chris (chris@nohup.cc) & Young (young@nohup.cc)
    """
    #print "start to read PSSM matrix"
    PSSM = []
    p = re.compile(r'-*[0-9]+')
    for line, strin in enumerate(input_matrix):
        if line > 2:
            str_vec = []
            overall_vec = strin.split()
            #print len(overall_vec)
            if len(overall_vec) == 0:
                break
            str_vec.extend(overall_vec[1])
            if(len(overall_vec) < 44):
                print "There is a mistake in the pssm file"
                print "Try to correct it"
                for cur_str in overall_vec[2:]:
                    str_vec.extend(p.findall(cur_str))
                    if(len(str_vec) >= 21):
                        if(len(str_vec)) >21:
                            #print len(str_vec)
                            #print str_vec
                            #print overall_vec
                            #print "Exit with an error"
                            exit(1)
                        break;
                print "Done"
            else:
                str_vec = strin.split()[1:42]
            if len(str_vec) == 0:
                break
            #str_vec_positive=map(int, str_vec[1:])
            PSSM.append(str_vec)
    fileinput.close()
    #print "finish to read PSSM matrix"
    PSSM = np.array(PSSM)
    return PSSM

def readPhycheValue(phyche_list_file, extra_index_file, k, alphabet, all_prop=False):

    phyche_list=""
    if all_prop == False and extra_index_file is None and phyche_list_file is None:
        if k == 2 and alphabet == Protein.index_list.DNA:
            print "parameters phyche_list_file is required. The default value of phyche_list is ['Rise', 'Roll', 'Shift', 'Slide', 'Tilt', 'Twist']."
            phyche_list=['Rise', 'Roll', 'Shift', 'Slide', 'Tilt', 'Twist']
        if k == 3 and alphabet == Protein.index_list.DNA:
            print "parameters phyche_list_file is required. The default value of phyche_list is ['Dnase I', 'Bendability (DNAse)']."
            phyche_list=['Dnase I', 'Bendability (DNAse)']
        if alphabet == Protein.index_list.PROTEIN:
            print "parameters phyche_list is required. The default value of phyche_list is ['Hydrophobicity', 'Hydrophilicity', 'Mass']."
            phyche_list=Protein.index_list.INDS_3_PROTEIN

    if phyche_list_file is not None:
        phyche_list = read_index(phyche_list_file)
    
    phyche_list = check_phyche_list(k, phyche_list, alphabet=alphabet, all_prop=all_prop)

    # Get phyche_vals.
    if alphabet == Protein.index_list.DNA or alphabet == Protein.index_list.RNA:
        if extra_index_file is not None:
            extra_phyche_index = get_extra_index(extra_index_file)
            phyche_vals = get_phyche_value(k, phyche_list, alphabet,
                                            normalize_index(extra_phyche_index, alphabet, is_convert_dict=True))
        else:
            phyche_vals = get_phyche_value(k, phyche_list, alphabet)
    elif alphabet == index_list.PROTEIN:
        phyche_vals = get_aaindex(phyche_list)
        if extra_index_file is not None:
            phyche_vals.extend(get_ext_ind_pro(extra_index_file))
        if k == 0:
            # Transform the data format to dict {acid: [phyche_vals]}.
            phyche_keys = phyche_vals[0].index_dict.keys()
            phyche_vals = [e.index_dict.values() for e in phyche_vals]
            new_phyche_vals = zip(*[e for e in phyche_vals])
            phyche_vals = {key: list(val) for key, val in zip(phyche_keys, new_phyche_vals)}
    return phyche_vals

def read_phyche_file_to_list(index_file):
    with open(index_file) as f_ind:
        lines = f_ind.readlines()
        ind_list = [index.rstrip() for index in lines]
        return ind_list

def check_phyche_list(k, phyche_list, alphabet, all_prop=False):
    """Get phyche_list and check it.

    :param k: int, the value of k-tuple.
    :param phyche_list: list, the input physicochemical properties list.
    :param all_prop: bool, choose all physicochemical properties or not.
    """

    # Set all_prop_list.
    all_prop_list = []
    try:
        if alphabet == index_list.DNA:
            if k == 2:
                all_prop_list = index_list.didna_list
            elif k == 3:
                all_prop_list = index_list.tridna_list
            else:
                error_info = 'Error, the k value must be 2 or 3.'
                raise ValueError(error_info)
        elif alphabet == index_list.RNA:
            if k == 2:
                all_prop_list = index_list.dirna_list
            else:
                error_info = 'Error, the k or alphabet error.'
                raise ValueError(error_info)
        elif alphabet == index_list.PROTEIN:
            all_prop_list = index_list.pro_list
        else:
            error_info = "Error, the alphabet must be dna, rna or protein."
            raise ValueError(error_info)
    except:
        raise

    # Set and check physicochemical properties.
    try:
        # Set all properties.
        if all_prop is True:
            phyche_list = all_prop_list
        # Check phyche properties.
        else:
            for e in phyche_list:
                if e not in all_prop_list:
                    error_info = 'Sorry, the physicochemical properties ' + e + ' is not exit.'
                    raise NameError(error_info)
    except:
        raise

    return phyche_list


def get_phyche_value(k, phyche_list, alphabet, extra_phyche_index=None):
    """Generate DNA or RNA phyche_value.

    :param k: int, the value of k-tuple.
    :param phyche_list: physicochemical properties list.
    :param extra_phyche_index: dict, the key is the olinucleotide (string),
                                     the value is its physicochemical property value (list).
                               It means the user-defined physicochemical indices.
    """
    if extra_phyche_index is None:
        extra_phyche_index = {}

    phyche_value = extend_phyche_index(get_phyche_index(k, phyche_list, alphabet), extra_phyche_index)

    return phyche_value

def extend_phyche_index(original_index, extend_index):
    """Extend DNA or RNA {phyche:[value, ... ]}"""
    if extend_index is None or len(extend_index) == 0:
        return original_index
    for key in list(original_index.keys()):
        original_index[key].extend(extend_index[key])
    return original_index


def get_phyche_index(k, phyche_list, alphabet):
    """get phyche_value according phyche_list."""
    phyche_value = {}
    if 0 == len(phyche_list):
        for nucleotide in make_kmer_list(k, alphabet):
            phyche_value[nucleotide] = []
        return phyche_value

    nucleotide_phyche_value = get_phyche_factor_dic(k, alphabet)
    for nucleotide in make_kmer_list(k, alphabet):
        if nucleotide not in phyche_value:
            phyche_value[nucleotide] = []
        for e in nucleotide_phyche_value[nucleotide]:
            if e[0] in phyche_list:
                phyche_value[nucleotide].append(e[1])

    return phyche_value

def get_phyche_factor_dic(k, alphabet):
    """Get all DNA or RNA {nucleotide: [(phyche, value), ...]} dict."""
    full_path = os.path.realpath(__file__)
    if 2 == k and alphabet == index_list.DNA:
        file_path = "%s/DNA/data/didna.data" % os.path.dirname(full_path)
    elif 2 == k and alphabet == index_list.RNA:
        file_path = "%s/RNA/data/dirna.data" % os.path.dirname(full_path)
    elif 3 == k:
        file_path = "%s/DNA/data/mmc4.data" % os.path.dirname(full_path)
    else:
        sys.stderr.write("The k can just be 2 or 3.")
        sys.exit(0)

    try:
        with open(file_path, 'rb') as f:
            phyche_factor_dic = pickle.load(f)
    except:
        with open(file_path, 'r') as f:
            phyche_factor_dic = pickle.load(f)

    return phyche_factor_dic

def get_aaindex(index_list):
    """Get the aaindex from data/aaindex.data.

    :param index_list: the index we want to get.
    :return: a list of AAIndex obj.
    """
    full_path = os.path.realpath(__file__)
    file_path = "%s/Protein/data/aaindex.data" % os.path.dirname(full_path)
    new_aaindex = []
    with open(file_path, 'rb') as f:
        aaindex = pickle.load(f)
        for index_vals in aaindex:
            if index_vals.head in index_list:
                new_aaindex.append(index_vals)

    return new_aaindex

def get_ext_ind_pro(filename):
    """Get the extend indices from index file, only work for protein."""
    # inds = ['A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y']
    aaindex = []
    with open(filename, 'r') as f:
        lines = f.readlines()
        for i, line in enumerate(lines):
            if line[0] == '>':
                temp_name = line[1:].rstrip()
                inds = lines[i+1].rstrip().split('\t')
                vals = lines[i+2].rstrip().split('\t')
                ind_val = {ind: float(val) for ind, val in zip(inds, vals)}
                aaindex.append(AAIndex(temp_name, ind_val))

    return aaindex

def normalize_index(phyche_index, alphabet, is_convert_dict=False):
    """Normalize the physicochemical index."""
    normalize_phyche_value = []
    for phyche_value in phyche_index:
        average_phyche_value = sum(phyche_value) * 1.0 / len(phyche_value)
        sd_phyche = standard_deviation(phyche_value)
        normalize_phyche_value.append([round((e - average_phyche_value) / sd_phyche, 2) for e in phyche_value])

    if is_convert_dict is True:
        return convert_phyche_index_to_dict(normalize_phyche_value, alphabet)

    print(normalize_phyche_value)
    return normalize_phyche_value

def standard_deviation(value_list):
    """Return standard deviation."""
    n = len(value_list)
    average_value = sum(value_list) * 1.0 / n
    return sqrt(sum([pow(e - average_value, 2) for e in value_list]) * 1.0 / (n - 1))

def convert_phyche_index_to_dict(phyche_index, alphabet):
    """Convert phyche index from list to dict."""
    # for e in phyche_index:
    #     print e
    len_index_value = len(phyche_index[0])
    k = 0
    for i in range(1, 10):
        if len_index_value < 4**i:
            sys.exit("Sorry, the number of each index value is must be 4^k.")
        if len_index_value == 4**i:
            k = i
            break
    kmer_list = make_kmer_list(k, alphabet)
    # print kmer_list
    len_kmer = len(kmer_list)
    phyche_index_dict = {}
    for kmer in kmer_list:
        phyche_index_dict[kmer] = []
    # print phyche_index_dict
    phyche_index = list(zip(*phyche_index))
    for i in range(len_kmer):
        phyche_index_dict[kmer_list[i]] = list(phyche_index[i])

    return phyche_index_dict

def get_extra_index(filename):
    """Get the extend indices from index file, only work for DNA and RNA."""
    extra_index_vals = []
    with open(filename) as f:
        lines = f.readlines()
        for ind, line in enumerate(lines):
            if line[0] == '>':
                vals = lines[ind + 2].rstrip().split('\t')
                vals = [float(val) for val in vals]
                extra_index_vals.append(vals)

    return extra_index_vals

def read_index(index_file):
    with open(index_file) as f_ind:
        lines = f_ind.readlines()
        ind_list = [index.rstrip() for index in lines]
        return ind_list