#! /usr/bin/env python
# -*- coding: utf-8 -*-
__author__ = 'Fule Liu'

import index_list,os,pickle,sys,re
from math import sqrt
from math import pow
import numpy as np
from itertools import combinations_with_replacement, permutations, product
from extract_aaindex import norm_index_vals, AAIndex
import math
import fileinput


def get_phyche_list(k, phyche_list, extra_index_file, alphabet, all_prop=False):
    """Get phyche_list and check it.

    :param k: int, the value of k-tuple.
    :param phyche_list: list, the input physicochemical properties list.
    :param all_prop: bool, choose all physicochemical properties or not.
    """
    if phyche_list is None or len(phyche_list) == 0:
        if extra_index_file is None and all_prop is False:
            error_info = 'Error, The phyche_list, extra_index_file and all_prop can\'t be all False.'
            raise ValueError(error_info)



    # Set all_prop_list.
    all_prop_list = []
    try:
        if alphabet == index_list.DNA:
            if k == 2:
                all_prop_list = index_list.didna_list
            elif k == 3:
                all_prop_list = index_list.tridna_list
            else:
                error_info = 'Error, the k value must be 2 or 3.'
                raise ValueError(error_info)
        elif alphabet == index_list.RNA:
            if k == 2:
                all_prop_list = index_list.dirna_list
            else:
                error_info = 'Error, the k or alphabet error.'
                raise ValueError(error_info)
        elif alphabet == index_list.PROTEIN:
            all_prop_list = index_list.pro_list
        else:
            error_info = "Error, the alphabet must be dna, rna or protein."
            raise ValueError(error_info)
    except:
        raise

    # Set and check physicochemical properties.
    try:
        # Set all properties.
        if all_prop is True:
            phyche_list = all_prop_list
        # Check phyche properties.
        else:
            for e in phyche_list:
                if e not in all_prop_list:
                    error_info = 'Sorry, the physicochemical properties ' + e + ' is not exit.'
                    raise NameError(error_info)
    except:
        raise

    return phyche_list


def get_extra_index(filename):
    """Get the extend indices from index file, only work for DNA and RNA."""
    extra_index_vals = []
    with open(filename) as f:
        lines = f.readlines()
        for ind, line in enumerate(lines):
            if line[0] == '>':
                vals = lines[ind + 2].rstrip().split('\t')
                vals = [float(val) for val in vals]
                extra_index_vals.append(vals)

    return extra_index_vals

def get_phyche_value(k, phyche_list, alphabet, extra_phyche_index=None):
    """Generate DNA or RNA phyche_value.

    :param k: int, the value of k-tuple.
    :param phyche_list: physicochemical properties list.
    :param extra_phyche_index: dict, the key is the olinucleotide (string),
                                     the value is its physicochemical property value (list).
                               It means the user-defined physicochemical indices.
    """
    if extra_phyche_index is None:
        extra_phyche_index = {}

    phyche_value = extend_phyche_index(get_phyche_index(k, phyche_list, alphabet), extra_phyche_index)

    return phyche_value


def get_phyche_index(k, phyche_list, alphabet):
    """get phyche_value according phyche_list."""
    phyche_value = {}
    if 0 == len(phyche_list):
        for nucleotide in make_kmer_list(k, alphabet):
            phyche_value[nucleotide] = []
        return phyche_value

    nucleotide_phyche_value = get_phyche_factor_dic(k, alphabet)
    for nucleotide in make_kmer_list(k, alphabet):
        if nucleotide not in phyche_value:
            phyche_value[nucleotide] = []
        for e in nucleotide_phyche_value[nucleotide]:
            if e[0] in phyche_list:
                phyche_value[nucleotide].append(e[1])

    return phyche_value

def get_phyche_factor_dic(k, alphabet):
    """Get all DNA or RNA {nucleotide: [(phyche, value), ...]} dict."""
    full_path = os.path.realpath(__file__)
    if 2 == k and alphabet == index_list.DNA:
        file_path = "%s/data/didna.data" % os.path.dirname(full_path)
    elif 2 == k and alphabet == index_list.RNA:
        file_path = "%s/data/dirna.data" % os.path.dirname(full_path)
    elif 3 == k:
        file_path = "%s/data/mmc4.data" % os.path.dirname(full_path)
    else:
        sys.stderr.write("The k can just be 2 or 3.")
        sys.exit(0)

    try:
        with open(file_path, 'rb') as f:
            phyche_factor_dic = pickle.load(f)
    except:
        with open(file_path, 'r') as f:
            phyche_factor_dic = pickle.load(f)

    return phyche_factor_dic


def extend_phyche_index(original_index, extend_index):
    """Extend DNA or RNA {phyche:[value, ... ]}"""
    if extend_index is None or len(extend_index) == 0:
        return original_index
    for key in list(original_index.keys()):
        original_index[key].extend(extend_index[key])
    return original_index

def make_kmer_list(k, alphabet):
    """
    Generate kmer list.
    """
    if k < 0:
        print("Error, k must be an inter and larger than 0.")

    kmers = []
    for i in range(1, k + 1):
        if len(kmers) == 0:
            kmers = list(alphabet)
        else:
            new_kmers = []
            for kmer in kmers:
                for c in alphabet:
                    new_kmers.append(kmer + c)
            kmers = new_kmers

    return kmers

def normalize_index(phyche_index, alphabet, is_convert_dict=False):
    """Normalize the physicochemical index."""
    normalize_phyche_value = []
    for phyche_value in phyche_index:
        average_phyche_value = sum(phyche_value) * 1.0 / len(phyche_value)
        sd_phyche = standard_deviation(phyche_value)
        normalize_phyche_value.append([round((e - average_phyche_value) / sd_phyche, 2) for e in phyche_value])

    if is_convert_dict is True:
        return convert_phyche_index_to_dict(normalize_phyche_value, alphabet)

    print(normalize_phyche_value)
    return normalize_phyche_value

def standard_deviation(value_list):
    """Return standard deviation."""
    n = len(value_list)
    average_value = sum(value_list) * 1.0 / n
    return sqrt(sum([pow(e - average_value, 2) for e in value_list]) * 1.0 / (n - 1))

def convert_phyche_index_to_dict(phyche_index, alphabet):
    """Convert phyche index from list to dict."""
    # for e in phyche_index:
    #     print e
    len_index_value = len(phyche_index[0])
    k = 0
    for i in range(1, 10):
        if len_index_value < 4**i:
            sys.exit("Sorry, the number of each index value is must be 4^k.")
        if len_index_value == 4**i:
            k = i
            break
    kmer_list = make_kmer_list(k, alphabet)
    # print kmer_list
    len_kmer = len(kmer_list)
    phyche_index_dict = {}
    for kmer in kmer_list:
        phyche_index_dict[kmer] = []
    # print phyche_index_dict
    phyche_index = list(zip(*phyche_index))
    for i in range(len_kmer):
        phyche_index_dict[kmer_list[i]] = list(phyche_index[i])

    return phyche_index_dict

def read_index(index_file):
    with open(index_file) as f_ind:
        lines = f_ind.readlines()
        ind_list = [index.rstrip() for index in lines]
        return ind_list

def getSpecificValue(olinuc, olinucs, prop, supInfo):
    # olinucs = getNucSeq(SupFileName).split(",")
    olinucs = olinucs.split(",")
    values = getValues(prop, supInfo).rstrip()
    values = values.split(",")
    # valueS = [float(x) for x in values.split(",")]
    count = olinucs.index(olinuc)
    value = values[count]
    return float(value)

def avgP(seq, olinucs, length, k, prop, supInfo):
    limit = length - k + 1
    i = 1
    sum = 0
    while i < limit or i == limit:
        # value = hn(seq[i - 1],prop,SupFileName)
        value = getSpecificValue(seq[i - 1], olinucs, prop, supInfo)
        sum = sum + value
        i = i + 1
    sum = sum / limit
    return sum

def getValues(prop, supInfo):
    values = ""
    name = re.search(prop, supInfo)
    if name:
        strr = prop + '\s*\,(.+)'
        b = re.search(strr, supInfo)
        if b:
            values = b.group(1)
    return values


def sepSequence(seq, k):
    i = k - 1
    seqq = []
    while i < len(seq):
        j = 0
        nuc = ''
        while j < k:
            nuc = seq[i - j] + nuc
            j = j + 1
        seqq.append(nuc)
        i += 1
    return seqq

def get_parallel_factor(k, lamada, sequence, phyche_value, alphabet):
    """Get the corresponding factor theta list."""
    theta = []
    l = len(sequence)

    for i in range(1, lamada + 1):
        temp_sum = 0.0
        for j in range(0, l - k - i + 1):
            nucleotide1 = sequence[j: j + k]
            nucleotide2 = sequence[j + i: j + i + k]
            if alphabet == index_list.DNA or alphabet == index_list.RNA:
                temp_sum += parallel_cor_function(nucleotide1, nucleotide2, phyche_value)
            elif alphabet == index_list.PROTEIN:
                temp_sum += pro_cor_fun1(nucleotide1, nucleotide2, phyche_value)

        theta.append(temp_sum / (l - k - i + 1))

    return theta

def parallel_cor_function(nucleotide1, nucleotide2, phyche_index):
    """Get the cFactor.(Type1)"""
    temp_sum = 0.0
    phyche_index_values = list(phyche_index.values())
    len_phyche_index = len(phyche_index_values[0])
    for u in range(len_phyche_index):
        temp_sum += pow(float(phyche_index[nucleotide1][u]) - float(phyche_index[nucleotide2][u]), 2)

    return temp_sum / len_phyche_index

def pro_cor_fun1(ri, rj, aaindex_list):
    _sum = 0.0
    len_index = len(aaindex_list)
    if isinstance(aaindex_list[0], tuple):
        for (head, index_dict) in aaindex_list:
            _sum += pow(index_dict[ri] - index_dict[rj], 2)
    else:
        for aaindex in aaindex_list:
            _sum += pow(aaindex.index_dict[ri] - aaindex.index_dict[rj], 2)
    return _sum / len_index

def frequency(tol_str, tar_str):
    """Generate the frequency of tar_str in tol_str.

    :param tol_str: mother string.
    :param tar_str: substring.
    """
    i, j, tar_count = 0, 0, 0
    len_tol_str = len(tol_str)
    len_tar_str = len(tar_str)
    while i < len_tol_str and j < len_tar_str:
        if tol_str[i] == tar_str[j]:
            i += 1
            j += 1
            if j >= len_tar_str:
                tar_count += 1
                i = i - j + 1
                j = 0
        else:
            i = i - j + 1
            j = 0

    return tar_count

def get_series_factor(k, lamada, sequence, phyche_value, alphabet):
    """Get the corresponding series factor theta list."""
    theta = []
    l_seq = len(sequence)
    if alphabet == index_list.DNA or alphabet == index_list.RNA:
        temp_values = list(phyche_value.values())
        max_big_lamada = len(temp_values[0])
    elif alphabet == index_list.PROTEIN:
        max_big_lamada = len(phyche_value)

    for small_lamada in range(1, lamada + 1):
        for big_lamada in range(max_big_lamada):
            temp_sum = 0.0
            for i in range(0, l_seq - k - small_lamada + 1):
                nucleotide1 = sequence[i: i + k]
                nucleotide2 = sequence[i + small_lamada: i + small_lamada + k]
                if alphabet == index_list.DNA or alphabet == index_list.RNA:
                    temp_sum += series_cor_function(nucleotide1, nucleotide2, big_lamada, phyche_value)
                elif alphabet == index_list.PROTEIN:
                    temp_sum += pro_cor_fun2(nucleotide1, nucleotide2, phyche_value[big_lamada])

            theta.append(temp_sum / (l_seq - k - small_lamada + 1))

    return theta

def series_cor_function(nucleotide1, nucleotide2, big_lamada, phyche_value):
    """Get the series correlation Factor(Type 2)."""
    return float(phyche_value[nucleotide1][big_lamada]) * float(phyche_value[nucleotide2][big_lamada])

def pro_cor_fun2(ri, rj, aaindex):
    return aaindex.index_dict[ri] * aaindex.index_dict[rj]

def get_kmer_lst(letterlst, k):
    """Generate a list of all possible k-mer pattern.

    :param letter: A list that contains all the possible letters in the sequence.
    :param k: The length of k-mer.
    :return: A kmer list.
    """
    kmerlst = []
    letter_set = set(letterlst)
    letter = [''.join(i) for i in letter_set]

    partkmers = list(combinations_with_replacement(letter, k))

    for element in partkmers:
        elelst = set(permutations(element, k))
        strlst = [''.join(ele) for ele in elelst]
        kmerlst += strlst
    kmerlst = np.sort(kmerlst)
    return list(kmerlst)

def delete_free_base(seqss):
    """Delete free base based on secondary structure to produce a new sequence and secondary structure. New sequence and secondary structure is a substring of the original sequence and secondary structure.
    :param seqss: a seqss object.
    :return: A new sequence and sstructure,string.
    """
    left_pos = seqss.sstructure.index('(')
    right_pos = seqss.sstructure.rindex(')')
    return seqss.sequence[left_pos:right_pos+1], seqss.sstructure[left_pos:right_pos+1]


def delete_loop(seqss):
    """Delete loop(hairpin) based on secondary structure to produce a new sequence and secondary structure. New sequence and secondary structure is a substring of the original sequence and secondary structure.

    :param seqss: a seqss object.
    :return: A new sequence and sstructure,string.
    """
    loop_re = r'(\(\.+\))'
    loop_list = re.findall(loop_re, seqss.sstructure)
    for loop in loop_list:
        pos = seqss.sstructure.index(loop)
        length = len(loop)
        sstructure_dl = seqss.sstructure[:pos+1] + seqss.sstructure[pos+length-1:]
        sequence_dl = seqss.sequence[:pos+1] + seqss.sequence[pos+length-1:]
    return sequence_dl, sstructure_dl

def get_corresp_sequence(seqss):
    sstructure_lst = list(seqss.sstructure)
    if len(seqss.sequence)==len(seqss.sstructure):
        s = []
        for i in range(len(seqss.sequence)):
            if seqss.sstructure[i] == "(":
                s.append(i)
            if seqss.sstructure[i] ==")":
                pos = s.pop()
                sstructure_lst[pos] = seqss.sequence[i]
                sstructure_lst[i] = seqss.sequence[pos]
        sstructure_new = ''.join(sstructure_lst)
        return sstructure_new
    else:
        error_info = seqss.name+'The length of sequence is not equal to the length of its secondary structure.'
        sys.stderr.write(error_info)

def get_pseknc_vector(seqss, k, letter_list = ['A', 'C', 'G', 'U', 'A-U', 'U-A', 'G-C', 'C-G', 'G-U', 'U-G']):
    '''This is a process in PseKNC, aim to generate feature vector.

    :param seqss: a seqss object.
    :param psekncdict: All the features, dictionary.
    :param k: The length of K-tuple nucleotide composition.
    :param letter_list: default   ['A', 'C', 'G', 'U', 'A-U', 'U-A', 'G-C', 'C-G', 'G-U', 'U-G'].
    :return: Feature vector through PseKNC.
    '''
    psekncdict = get_pseknc_dict(letter_list, k)
    vector = np.zeros((1, len(psekncdict)))
    correspseq = get_corresp_sequence(seqss)
    pattern = zip(list(seqss.sequence),list(correspseq))

    for i in xrange(len(pattern)-k+1):
        stem = []
        for x, y in pattern[i:i+k]:
            if x == '.' or y == '.':
                if x == '.':
                    stem.append(y)
                else:
                    stem.append(x)
            else:
                stem.append(x + '-' + y)
        stem_tuple= tuple(stem)
        position = psekncdict.get(stem_tuple)
        vector[0, position] += 1
        #print stem_tuple,position
    #return vector[0]
    return list(vector[0]/sum(vector[0]))
    #return [round(f,4) for f in list(vector[0]/sum(vector[0]))]

def calculate_theta(seqss, j):
    '''calculate theta
    :param seqss: a seqss object.
    :param j: the counted rank (or tier) of the structural correlation along a RNA chain.
    :return: theta.
    '''
    if j >= len(seqss.sequence):
        error_info = 'Error occured in '+seqss.id +', r should be less than the length of the sequence.'
        sys.stderr.write(error_info)
    else:
        correspseq = get_corresp_sequence(seqss)
        pattern = zip(list(seqss.sequence), list(correspseq))
        stem=[]
        for x,y in pattern:
            if x == '.' or y == '.':
                if x == '.':
                    stem.append(y)
                else:
                    stem.append(x)
            else:
                stem.append(x + '-' + y)
        freevalue_vector = []
        for i in stem:
            if i == 'A-U' or i == 'U-A':
               freevalue_vector.append(-2)
            elif i == 'C-G' or i == 'G-C':
                freevalue_vector.append(-3)
            elif i == 'U-G' or i == 'G-U':
                freevalue_vector.append(-1)
            else:
                freevalue_vector.append(0)

        s=0.0
        for i in range(len(freevalue_vector)-j):
            s += (freevalue_vector[i] - freevalue_vector[i+j]) ** 2
            #print i,i+j
            #print (freevalue_vector[i] - freevalue_vector[i+r])
        #print s,len(freevalue_vector)-r
        return s / (len(freevalue_vector)-j)

def get_pseknc_dict(letter_list, k):
    """Generate a dictionary of all possible PseKNC pattern.
    :param letter: A list that contains all the possible characters in an RNA sequence. eg:['A','C','G','U']
    :param k: The length of K-tuple nucleotide composition.
    :return: A PseKNC pattern dictionary.
    """
    pseknclst = []
    part_psessc = list(combinations_with_replacement(letter_list, k))
    for element in part_psessc:
        elelst = set(permutations(element, k))
        pseknclst += elelst
    pseknclst.sort()
    psekncdict = {pseknclst[i]:i for i in range(len(pseknclst))}
    return psekncdict


def get_aaindex(index_list):
    """Get the aaindex from data/aaindex.data.

    :param index_list: the index we want to get.
    :return: a list of AAIndex obj.
    """
    full_path = os.path.realpath(__file__)
    file_path = "%s/data/aaindex.data" % os.path.dirname(full_path)
    new_aaindex = []
    with open(file_path, 'rb') as f:
        aaindex = pickle.load(f)
        for index_vals in aaindex:
            if index_vals.head in index_list:
                new_aaindex.append(index_vals)

    return new_aaindex


def get_ext_ind_pro(filename):
    """Get the extend indices from index file, only work for protein."""
    # inds = ['A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y']
    aaindex = []
    with open(filename, 'r') as f:
        lines = f.readlines()
        for i, line in enumerate(lines):
            if line[0] == '>':
                temp_name = line[1:].rstrip()
                inds = lines[i+1].rstrip().split('\t')
                vals = lines[i+2].rstrip().split('\t')
                ind_val = {ind: float(val) for ind, val in zip(inds, vals)}
                aaindex.append(AAIndex(temp_name, ind_val))

    return aaindex

def checkFasta(fastas):
	status = True
	lenList = set()
	for i in fastas:
		lenList.add(len(i[1]))
	if len(lenList) == 1:
		return True
	else:
		return False

def minSequenceLength(fastas):
	minLen = 10000
	for i in fastas:
		if minLen > len(i[1]):
			minLen = len(i[1])
	return minLen

def minSequenceLengthWithNormalAA(fastas):
	minLen = 10000
	for i in fastas:
		if minLen > len(re.sub('-', '', i[1])):
			minLen = len(re.sub('-', '', i[1]))
	return minLen

def readFasta(file):
	if os.path.exists(file) == False:
		print('Error: "' + file + '" does not exist.')
		sys.exit(1)

	with open(file) as f:
		records = f.read()

	if re.search('>', records) == None:
		print('The input file seems not in fasta format.')
		sys.exit(1)
    	records = records.replace('\r','').split('>')[1:]
    	myFasta = []
    	for fasta in records:
        	array = fasta.split('\n')
        	name, sequence = array[0].split()[0], re.sub('[^ARNDCQEGHILKMFPSTWYVU-]', '-', ''.join(array[1:]).upper())
        	myFasta.append([name, sequence])
    	return myFasta

# Authors: Superzchen
def gapModel(fastas, myDict, gDict, gNames, ktuple, glValue):
    encodings = []

    if ktuple == 1:
        for i in fastas:
            name, sequence = i[0], re.sub('-', '', i[1])
            code = []
            numDict = {}
            for j in range(0, len(sequence), glValue+1):
                numDict[gDict[myDict[sequence[j]]]] = numDict.get(gDict[myDict[sequence[j]]], 0) + 1

            for g in gNames:
                code.append(numDict.get(g, 0))
            encodings.append(code)

    if ktuple == 2:
        for i in fastas:
            name, sequence = i[0], re.sub('-', '', i[1])
            code = []
            numDict = {}
            for j in range(0, len(sequence), glValue + 1):
                if j+1 < len(sequence):
                    numDict[gDict[myDict[sequence[j]]]+'_'+gDict[myDict[sequence[j+1]]]] = numDict.get(gDict[myDict[sequence[j]]]+'_'+gDict[myDict[sequence[j+1]]], 0) + 1

            for g in [g1+'_'+g2 for g1 in gNames for g2 in gNames]:
                code.append(numDict.get(g, 0))
            encodings.append(code)

    if ktuple == 3:
        for i in fastas:
            name, sequence = i[0], re.sub('-', '', i[1])
            code = []
            numDict = {}
            for j in range(0, len(sequence), glValue + 1):
                if j + 1 < len(sequence) and j + 2 < len(sequence):
                    numDict[gDict[myDict[sequence[j]]] + '_' + gDict[myDict[sequence[j + 1]]] + '_' + gDict[myDict[sequence[j + 2]]]] = numDict.get(gDict[myDict[sequence[j]]] + '_' + gDict[myDict[sequence[j + 1]]] + '_' + gDict[myDict[sequence[j + 2]]], 0) + 1

            for g in [g1 + '_' + g2 + '_' +g3 for g1 in gNames for g2 in gNames for g3 in gNames]:
                code.append(numDict.get(g, 0))
            encodings.append(code)

    return encodings

# Authors: Superzchen
def lambdaModel(fastas, myDict, gDict, gNames, ktuple, glValue):
    if glValue == 0:
        print('Warning: the lambda value should not be zero in "lambda correlation" model')
        return 0

    encodings = []

    if ktuple == 1:
        for i in fastas:
            name, sequence = i[0], re.sub('-', '', i[1])
            code = []
            numDict = {}
            for j in range(0, len(sequence)):
                numDict[gDict[myDict[sequence[j]]]] = numDict.get(gDict[myDict[sequence[j]]], 0) + 1

            for g in gNames:
                code.append(numDict.get(g, 0))
            encodings.append(code)

    if ktuple == 2:
        for i in fastas:
            name, sequence = i[0], re.sub('-', '', i[1])
            code = []
            numDict = {}
            for j in range(0, len(sequence)):
                if j + glValue < len(sequence):
                    numDict[gDict[myDict[sequence[j]]] + '_' + gDict[myDict[sequence[j + glValue]]]] = numDict.get(
                        gDict[myDict[sequence[j]]] + '_' + gDict[myDict[sequence[j + glValue]]], 0) + 1

            for g in [g1 + '_' + g2 for g1 in gNames for g2 in gNames]:
                code.append(numDict.get(g, 0))
            encodings.append(code)

    if ktuple == 3:
        for i in fastas:
            name, sequence = i[0], re.sub('-', '', i[1])
            code = []
            numDict = {}
            for j in range(0, len(sequence)):
                if j + glValue < len(sequence) and j + 2*glValue < len(sequence):
                    numDict[gDict[myDict[sequence[j]]] + '_' + gDict[myDict[sequence[j + glValue]]] + '_' + gDict[
                        myDict[sequence[j + 2*glValue]]]] = numDict.get(
                        gDict[myDict[sequence[j]]] + '_' + gDict[myDict[sequence[j + glValue]]] + '_' + gDict[
                            myDict[sequence[j + 2*glValue]]], 0) + 1

            for g in [g1 + '_' + g2 + '_' + g3 for g1 in gNames for g2 in gNames for g3 in gNames]:
                code.append(numDict.get(g, 0))
            encodings.append(code)

    return encodings

def readToMatrix(input_matrix):
    """
    Authors: Chris (chris@nohup.cc) & Young (young@nohup.cc)
    """
    #print "start to read PSSM matrix"
    PSSM = []
    p = re.compile(r'-*[0-9]+')
    for line, strin in enumerate(input_matrix):
        if line > 2:
            str_vec = []
            overall_vec = strin.split()
            #print len(overall_vec)
            if len(overall_vec) == 0:
                break
            str_vec.extend(overall_vec[1])
            if(len(overall_vec) < 44):
                print "There is a mistake in the pssm file"
                print "Try to correct it"
                for cur_str in overall_vec[2:]:
                    str_vec.extend(p.findall(cur_str))
                    if(len(str_vec) >= 21):
                        if(len(str_vec)) >21:
                            #print len(str_vec)
                            #print str_vec
                            #print overall_vec
                            #print "Exit with an error"
                            exit(1)
                        break;
                print "Done"
            else:
                str_vec = strin.split()[1:42]
            if len(str_vec) == 0:
                break
            #str_vec_positive=map(int, str_vec[1:])
            PSSM.append(str_vec)
    fileinput.close()
    #print "finish to read PSSM matrix"
    PSSM = np.array(PSSM)
    return PSSM

def handleRows(PSSM, SWITCH, COUNT):
    '''
    Authors: Chris (chris@nohup.cc) & Young (young@nohup.cc)
    if SWITCH=0, we filter no element.
    if SWITCH=1, we filter all the negative elements.
    if SWITCH=2, we filter all the negative and positive elements greater than expected.
    '''
    '''
    if COUNT=20, we generate a 20-dimension vector.
    if COUNT=400, we generate a 400-dimension vector.
    '''
    # 0-19 represents amino acid 'ARNDCQEGHILKMFPSTWYV'
    Amino_vec = "ARNDCQEGHILKMFPSTWYV"

    matrix_final = [ [0.0] * 20 ] * (COUNT/20)
    matrix_final=np.array(matrix_final)
    seq_cn = 0

    PSSM_shape=np.shape(PSSM)
    for i in range(PSSM_shape[0]):
        seq_cn += 1
        str_vec=PSSM[i]
        str_vec_positive=map(int, str_vec[1:21])
        str_vec_positive=np.array(str_vec_positive)
        if SWITCH==1:
            str_vec_positive[str_vec_positive<0]=0
        elif SWITCH==2:
            str_vec_positive[str_vec_positive<0]=0
            str_vec_positive[str_vec_positive>7]=0
        #print "str_vec_positive="
        #print str_vec_positive
        if COUNT==20:
            matrix_final[0]=map(sum, zip(str_vec_positive, matrix_final[0]))
        elif COUNT==400:
            matrix_final[Amino_vec.index(str_vec[0])] = map(sum, zip(str_vec_positive, matrix_final[Amino_vec.index(str_vec[0])]))

        #print "matrix_final="
        #print matrix_final

    return matrix_final

def average(matrixSum, seqLen):
    '''Authors: Chris (chris@nohup.cc) & Young (young@nohup.cc)'''
    # average the summary of rows
    matrix_array = np.array(matrixSum)
    matrix_array = np.divide(matrix_array, seqLen)
    matrix_array_shp = np.shape(matrix_array)
    matrix_average = [(np.reshape(matrix_array, (matrix_array_shp[0] * matrix_array_shp[1], )))]
    return matrix_average

def window(PSSM, w_smth, w_slide):
    '''Authors: Chris (chris@nohup.cc) & Young (young@nohup.cc)'''
    # 0-19 represents amino acid 'ARNDCQEGHILKMFPSTWYV'
    w_smth=int(w_smth)
    w_slide=int(w_slide)
    Amino_vec = "ARNDCQEGHILKMFPSTWYV"

    PSSM=PSSM[:,1:21]
    PSSM=PSSM.astype(float)

    seq_cn = np.shape(PSSM)[0]

    #original PSSM
    PSSM_smth = np.array([[0.0]*20]*seq_cn)
    PSSM_orig = np.array(PSSM)
    #print PSSM_orig
    #section for PSSM_smth features
    PSSM_smth_full=pssm_smth(PSSM_orig,PSSM_smth,w_smth,seq_cn)
    #print PSSM_smth_full
    PSSM_smth_final=[[0.0]*20]*w_slide
    PSSM_smth_final=np.array(PSSM_smth_final)
    for i in range(w_slide):
        PSSM_smth_final[i]=PSSM_smth_full[i]
    matrix_final=average(PSSM_smth_final,1.0)
    return matrix_final

def pssm_smth(PSSM_orig, PSSM_smth, w_smth, l):
    '''Authors: Chris (chris@nohup.cc) & Young (young@nohup.cc)'''
    for i in range(l):
        if i <(w_smth-1)/2:
            for j in range(i+(w_smth-1)/2+1):
                PSSM_smth[i]+=PSSM_orig[j]
        elif i>=(l-(w_smth-1)/2):
            for j in range(i-(w_smth-1)/2,l):
                PSSM_smth[i]+=PSSM_orig[j]
        else:
            for j in range(i-(w_smth-1)/2,i+(w_smth-1)/2+1):
                PSSM_smth[i]+=PSSM_orig[j]
    return PSSM_smth

def preHandleColumns(PSSM,STEP,PART,ID):
    '''Authors: Chris (chris@nohup.cc) & Young (young@nohup.cc)'''
    '''
    if STEP=k, we calculate the relation betweem one residue and the kth residue afterward.
    '''
    '''
    if PART=0, we calculate the left part of PSSM.
    if PART=1, we calculate the right part of PSSM.
    '''
    '''
    if ID=0, we product the residue-pair.
    if ID=1, we minus the residue-pair.
    '''
    '''
    if KEY=1, we divide each element by the sum of elements in its column.
    if KEY=0, we don't perform the above process.
    '''
    if PART==0:
        #print "PART=",PART
        PSSM=PSSM[:,1:21]
    elif PART==1:
        #print "PART=",PART
        PSSM=PSSM[:, 21:]
    PSSM=PSSM.astype(float)
    matrix_final = [ [0.0] * 20 ] * 20
    matrix_final=np.array(matrix_final)
    seq_cn=np.shape(PSSM)[0]
    #print "seq_cn=",seq_cn

    if ID==0:
        #print "ID=",ID
        for i in range(20):
            for j in range(20):
                for k in range(seq_cn - STEP):
                    matrix_final[i][j]+=(PSSM[k][i]*PSSM[k+STEP][j])

    elif ID==1:
        #print "ID=",ID
        for i in range(20):
            for j in range(20):
                for k in range(seq_cn - STEP):
                    matrix_final[i][j] += ((PSSM[k][i]-PSSM[k+STEP][j]) * (PSSM[k][i]-PSSM[k+STEP][j])/4.0)
    #print matrix_final
    return matrix_final

def handleTriColumns(PSSM):
    '''Authors: Chris (chris@nohup.cc) & Young (young@nohup.cc)'''
    matrix_final=[ [ [0.0] * 20] * 20] * 20
    matrix_final=np.array(matrix_final)
    PSSM=PSSM[:, 21:]
    PSSM=PSSM.astype(float)
    PSSM=np.array(PSSM)
    #print "PSSM="
    #print PSSM
    seq_cn=np.shape(PSSM)[0]
    for m in range(20):
        for n in range(20):
            for r in range(20):
                for i in range(seq_cn-2):
                    matrix_final[m][n][r]+=(PSSM[i][m]*PSSM[i+1][n]*PSSM[i+2][r])
    #print "matrix_final="
    #print matrix_final
    matrix_final=np.divide(matrix_final,1000000.0)
    matrix_final_shape=np.shape(matrix_final)
    #print matrix_final
    matrix_result = [(np.reshape(matrix_final, (matrix_final_shape[0] * matrix_final_shape[1] * matrix_final_shape[2], )))]
    return matrix_result

def handleMixed(PSSM,ALPHA):
    '''Authors: Chris (chris@nohup.cc) & Young (young@nohup.cc)'''
    row1=[0.0] * 20
    row2=[0.0] * 20

    matrix_final = [ [0.0] * 40 ] * 1
    row1=np.array(row1)
    row2=np.array(row2)
    matrix_final=np.array(matrix_final)

    PSSM_norm=normalizePSSM(PSSM)
    seq_cn=np.shape(PSSM)[0]
    for i in range(seq_cn):
        #print PSSM_norm[i]
        row1=map(sum, zip(row1, PSSM_norm[i]))
    #print row1
    row1=np.divide(row1,seq_cn)

    for j in range(20):
        for i in range(seq_cn-ALPHA):
            row2[j]+=(PSSM_norm[i][j]-PSSM_norm[i+ALPHA][j])*(PSSM_norm[i][j]-PSSM_norm[i+ALPHA][j])
    #print row2
    row2=np.divide(row2,seq_cn-ALPHA)

    row=np.hstack((row1,row2))
    matrix_final[0]=row
    #print np.shape(matrix_final)
    return matrix_final

def handleMixed2(PSSM,ALPHA):
    '''Authors: Chris (chris@nohup.cc) & Young (young@nohup.cc)'''
    row1=[0.0] * 40
    row2=[ [0.0] * (2*ALPHA) ]*20
    matrix_final = [ [0.0] * (40+40*ALPHA) ] * 1

    row1=np.array(row1)
    row2=np.array(row2)
    matrix_final=np.array(matrix_final)

    PSSM_norm=normalizePSSM(PSSM)
    seq_cn=np.shape(PSSM)[0]
    for j in range(20):
        positive_count_1=0
        negative_count_1=0
        for i in range(seq_cn):
            if PSSM_norm[i][j]>=0:

                positive_count_1+=1
                row1[2*j]+=PSSM_norm[i][j]
            elif PSSM_norm[i][j]<0:

                negative_count_1+=1
                row1[2*j+1]+=PSSM_norm[i][j]
        #print "positive_count_1="
        #print positive_count_1
        #print "negative_count_1="
        #print negative_count_1
        row1[2*j]=row1[2*j]/positive_count_1
        row1[2*j+1]=row1[2*j+1]/negative_count_1

    for j in range(20):
        for alpha in range(1,ALPHA+1):
            positive_count_2=0
            negative_count_2=0
            for i in range(seq_cn-alpha):
                if (PSSM_norm[i][j]-PSSM_norm[i+alpha][j])>=0:
                    positive_count_2+=1
                    row2[j][2*alpha-2]+=(PSSM_norm[i][j]-PSSM_norm[i+alpha][j])*(PSSM_norm[i][j]-PSSM_norm[i+alpha][j])
                elif (PSSM_norm[i][j]-PSSM_norm[i+alpha][j])<0:
                    negative_count_2+=1
                    row2[j][2*alpha-1]+=(PSSM_norm[i][j]-PSSM_norm[i+alpha][j])*(PSSM_norm[i][j]-PSSM_norm[i+alpha][j])
            row2[j][2*alpha-2]=row2[j][2*alpha-2]/positive_count_2
            row2[j][2*alpha-1]=row2[j][2*alpha-1]/negative_count_2

    row2=average(row2,1.0)
    row=np.hstack((row1,row2[0]))
    matrix_final[0]=row
    #print np.shape(matrix_final)
    return matrix_final

def handleMixed3(PSSM):
    '''Authors: Chris (chris@nohup.cc) & Young (young@nohup.cc)'''
    row=[ [0.0] * 110 ]*1
    row1=[0.0] * 100
    row2=[0.0] * 10
    row=np.array(row)
    row1=np.array(row1)
    row2=np.array(row2)

    seq_cn=np.shape(PSSM)[0]
    Amino_vec = "ARNDCQEGHILKMFPSTWYV"
    RPSSM=[ [0.0]*10 ]*seq_cn
    RPSSM=np.array(RPSSM)

    PSSM=PSSM[:,1:21]
    PSSM=PSSM.astype(float)
    PSSM=np.array(PSSM)
    #print "PSSM="
    #print PSSM
    RPSSM[:,0]=np.divide(map(sum,zip(PSSM[:,13],PSSM[:,17],PSSM[:,18])), 3.0)
    RPSSM[:,1]=np.divide(map(sum,zip(PSSM[:,10],PSSM[:,12])), 2.0)
    RPSSM[:,2]=np.divide(map(sum,zip(PSSM[:,9],PSSM[:,19])), 2.0)
    RPSSM[:,3]=np.divide(map(sum,zip(PSSM[:,0],PSSM[:,15],PSSM[:,16])), 3.0)
    RPSSM[:,4]=np.divide(map(sum,zip(PSSM[:,2],PSSM[:,8])), 2.0)
    RPSSM[:,5]=np.divide(map(sum,zip(PSSM[:,5],PSSM[:,6],PSSM[:,3])), 3.0)
    RPSSM[:,6]=np.divide(map(sum,zip(PSSM[:,1],PSSM[:,11])), 2.0)
    RPSSM[:,7]=PSSM[:,4]
    RPSSM[:,8]=PSSM[:,7]
    RPSSM[:,9]=PSSM[:,14]
    # print "RPSSM="
    # print RPSSM
    mean_matrix=np.mean(RPSSM, axis=0)
    # print "mean_matrix="
    # print mean_matrix
    for j in range(10):
        for i in range(seq_cn):
            row2[j]+=(RPSSM[i][j]-mean_matrix[j])*(RPSSM[i][j]-mean_matrix[j])
    #print "row2="
    #print row2
    row2=np.divide(row2,seq_cn)
    matrix_final = [ [0.0] * 10 ] * 10
    matrix_final=np.array(matrix_final)
    for i in range(10):
        for j in range(10):
            for k in range(seq_cn-1):
                matrix_final[i][j] += ((RPSSM[k][i]-RPSSM[k+1][j]) * (RPSSM[k][i]-RPSSM[k+1][j])/2.0)
    #print "matrix_final="
    #print matrix_final
    row1=average(matrix_final,seq_cn-1)[0]
    #print "row1="
    #print row1
    row[0]=np.hstack((row1,row2))
    return row

def normalizePSSM(PSSM):
    '''Authors: Chris (chris@nohup.cc) & Young (young@nohup.cc)'''
    PSSM=PSSM[:,1:21]
    PSSM=PSSM.astype(float)
    PSSM=np.array(PSSM)
    seq_cn=np.shape(PSSM)[0]
    PSSM_norm=[ [0.0] * 20 ] * seq_cn
    PSSM_norm=np.array(PSSM_norm)
    mean_matrix=np.mean(PSSM, axis=1)
    std_matrix=np.std(PSSM, axis=1)

    for i in range(seq_cn):
        for j in range(20):
            if std_matrix[i]==0.0:
                PSSM_norm[i][j] = PSSM[i][j]-mean_matrix[i]
            else:
                PSSM_norm[i][j]=(PSSM[i][j]-mean_matrix[i])/std_matrix[i]
    return PSSM_norm

def correlation(PSSM,ID,GROUP):
    '''Authors: Chris (chris@nohup.cc) & Young (young@nohup.cc)'''
    # 0-19 represents amino acid 'ARNDCQEGHILKMFPSTWYV'
    Amino_vec = "ARNDCQEGHILKMFPSTWYV"
    #GROUP=10
    PSSM=PSSM[:,1:21]
    PSSM=PSSM.astype(float)
    #print PSSM
    #section for PSSM_AC features
    seq_cn = np.shape(PSSM)[0]
    g=GROUP
    l=seq_cn
    if ID==0:
        matrix_final=pssm_ac_cal(PSSM,g,l)
    elif ID==1:
        matrix_final=pssm_cc_cal(PSSM,g,l)

    matrix_final =  average(matrix_final, l)
    return matrix_final

def pssm_ac_cal(PSSM, g, l):
    '''Authors: Chris (chris@nohup.cc) & Young (young@nohup.cc)'''
    PSSM_AC = np.array([ [0.0] * 20 ] * g)

    for pg in range(g):
        l_g = l - pg - 1
        for pj in range(20):
            sum_jl = 0.0
            for i in range(l):
                sum_jl += PSSM[i][pj]
            sum_jl /= l

            pssm_acjg = 0.0
            for i in range(l_g):
                pssm_acjg +=  (PSSM[i][pj]-sum_jl) * (PSSM[i+pg+1][pj]-sum_jl)
            pssm_acjg /= l_g
            PSSM_AC[pg][pj] = pssm_acjg
    return PSSM_AC

def pssm_cc_cal(PSSM, g, l):
    '''Authors: Chris (chris@nohup.cc) & Young (young@nohup.cc)'''
    PSSM_CC = np.array([ [0.0] * 380 ] * g)
    for pg in range(g):
        l_g = l - pg - 1
        for pj_1 in range(20):
            sum_jl_1 = 0.0
            for i in range(l):
                sum_jl_1 += PSSM[i][pj_1]
            sum_jl_1 /= l

            for pj_2 in range(20):
                if pj_2!=pj_1:
                    sum_jl_2 = 0.0
                    for i in range(l):
                        sum_jl_2 += PSSM[i][pj_2]
                    sum_jl_2 /= l
                    pssm_acjg = 0.0
                    for i in range(l_g):
                        pssm_acjg += (PSSM[i][pj_1]-sum_jl_1) * (PSSM[i+pg+1][pj_2]-sum_jl_2)
                    pssm_acjg /= l_g
                    if(pj_1<pj_2):
                        PSSM_CC[pg][19*pj_1+(pj_2-1)] = pssm_acjg
                    else:
                        PSSM_CC[pg][19*pj_1+(pj_2)] = pssm_acjg

    return PSSM_CC